"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var pluralize = require("pluralize");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var TypeNames_1 = require("./TypeNames");
var UnifyClasses_1 = require("./UnifyClasses");
var TypeNames_2 = require("./TypeNames");
var TypeAttributes_1 = require("./TypeAttributes");
var PathElementKind;
(function (PathElementKind) {
    PathElementKind[PathElementKind["Root"] = 0] = "Root";
    PathElementKind[PathElementKind["Definition"] = 1] = "Definition";
    PathElementKind[PathElementKind["OneOf"] = 2] = "OneOf";
    PathElementKind[PathElementKind["AnyOf"] = 3] = "AnyOf";
    PathElementKind[PathElementKind["Property"] = 4] = "Property";
    PathElementKind[PathElementKind["AdditionalProperty"] = 5] = "AdditionalProperty";
    PathElementKind[PathElementKind["Items"] = 6] = "Items";
})(PathElementKind = exports.PathElementKind || (exports.PathElementKind = {}));
function checkStringArray(arr) {
    if (!Array.isArray(arr)) {
        return Support_1.panic("Expected a string array, but got " + arr);
    }
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var e = arr_1[_i];
        if (typeof e !== "string") {
            return Support_1.panic("Expected string, but got " + e);
        }
    }
    return arr;
}
function parseRef(ref) {
    if (typeof ref !== "string") {
        return Support_1.panic("$ref must be a string");
    }
    var refName = "Something";
    var parts = ref.split("/");
    var elements = [];
    for (var i = 0; i < parts.length; i++) {
        if (parts[i] === "#") {
            elements.push({ kind: PathElementKind.Root });
            refName = "Root";
        }
        else if (parts[i] === "definitions" && i + 1 < parts.length) {
            refName = parts[i + 1];
            elements.push({ kind: PathElementKind.Definition, name: refName });
            i += 1;
        }
        else {
            Support_1.panic("Could not parse JSON schema reference " + ref);
        }
    }
    return [immutable_1.List(elements), refName];
}
function lookupDefinition(schema, name) {
    var definitions = Support_1.checkStringMap(schema.definitions);
    return Support_1.checkStringMap(definitions[name]);
}
function lookupProperty(schema, name) {
    var properties = Support_1.checkStringMap(schema.properties);
    return Support_1.checkStringMap(properties[name]);
}
function indexArray(cases, index) {
    if (!Array.isArray(cases)) {
        return Support_1.panic("oneOf or anyOf value must be an array");
    }
    return Support_1.checkStringMap(cases[index]);
}
function makeAttributes(schema, attributes) {
    var maybeDescription = schema.description;
    if (typeof maybeDescription === "string") {
        attributes = TypeAttributes_1.descriptionTypeAttributeKind.setInAttributes(attributes, maybeDescription);
    }
    return TypeNames_2.modifyTypeNames(attributes, function (maybeTypeNames) {
        var typeNames = Support_1.defined(maybeTypeNames);
        if (!typeNames.areInferred) {
            return typeNames;
        }
        var title = schema.title;
        if (typeof title === "string") {
            return new TypeNames_1.TypeNames(immutable_1.OrderedSet([title]), immutable_1.OrderedSet(), false);
        }
        else {
            return typeNames.makeInferred();
        }
    });
}
function checkTypeList(typeOrTypes) {
    if (typeof typeOrTypes === "string") {
        return immutable_1.OrderedSet([typeOrTypes]);
    }
    else if (Array.isArray(typeOrTypes)) {
        var arr = [];
        for (var _i = 0, typeOrTypes_1 = typeOrTypes; _i < typeOrTypes_1.length; _i++) {
            var t = typeOrTypes_1[_i];
            if (typeof t !== "string") {
                return Support_1.panic("element of type is not a string: " + t);
            }
            arr.push(t);
        }
        var set = immutable_1.OrderedSet(arr);
        Support_1.assert(!set.isEmpty(), "JSON Schema must specify at least one type");
        return set;
    }
    else {
        return Support_1.panic("type is neither a string or array of strings: " + typeOrTypes);
    }
}
function makeImmutablePath(path) {
    return path.map(function (pe) { return immutable_1.fromJS(pe); });
}
exports.rootRef = immutable_1.List([{ kind: PathElementKind.Root }]);
function addTypesInSchema(typeBuilder, rootJson, conflateNumbers, references) {
    var root = Support_1.checkStringMap(rootJson);
    var typeForPath = immutable_1.Map();
    function setTypeForPath(path, t) {
        typeForPath = typeForPath.set(makeImmutablePath(path), t);
    }
    function lookupRef(local, localPath, ref) {
        var first = ref.first();
        if (first === undefined) {
            return [local, localPath];
        }
        var rest = ref.rest();
        if (first.kind === PathElementKind.Root) {
            return lookupRef(root, immutable_1.List([first]), ref.rest());
        }
        localPath = localPath.push(first);
        switch (first.kind) {
            case PathElementKind.Definition:
                return lookupRef(lookupDefinition(local, first.name), localPath, rest);
            case PathElementKind.OneOf:
                return lookupRef(indexArray(local.oneOf, first.index), localPath, rest);
            case PathElementKind.AnyOf:
                return lookupRef(indexArray(local.anyOf, first.index), localPath, rest);
            case PathElementKind.Property:
                return lookupRef(lookupProperty(local, first.name), localPath, rest);
            case PathElementKind.AdditionalProperty:
                return lookupRef(Support_1.checkStringMap(local.additionalProperties), localPath, rest);
            case PathElementKind.Items:
                return lookupRef(Support_1.checkStringMap(local.items), localPath, rest);
            default:
                return Support_1.assertNever(first);
        }
    }
    function makeClass(path, attributes, properties, requiredArray) {
        var required = immutable_1.Set(requiredArray);
        var propertiesMap = immutable_1.Map(properties);
        var propertyDescriptions = propertiesMap
            .map(function (propSchema) {
            if (typeof propSchema === "object") {
                var desc = propSchema.description;
                if (typeof desc === "string") {
                    return desc;
                }
            }
            return undefined;
        })
            .filter(function (v) { return v !== undefined; });
        if (!propertyDescriptions.isEmpty()) {
            attributes = TypeAttributes_1.propertyDescriptionsTypeAttributeKind.setInAttributes(attributes, propertyDescriptions);
        }
        var result = typeBuilder.getUniqueClassType(attributes, true);
        setTypeForPath(path, result);
        // FIXME: We're using a Map instead of an OrderedMap here because we represent
        // the JSON Schema as a JavaScript object, which has no map ordering.  Ideally
        // we would use a JSON parser that preserves order.
        var props = propertiesMap.map(function (propSchema, propName) {
            var t = toType(Support_1.checkStringMap(propSchema), path.push({ kind: PathElementKind.Property, name: propName }), TypeNames_2.makeNamesTypeAttributes(pluralize.singular(propName), true));
            var isOptional = !required.has(propName);
            return new Type_1.ClassProperty(t, isOptional);
        });
        typeBuilder.setClassProperties(result, props.toOrderedMap());
        return result;
    }
    function makeMap(path, typeAttributes, additional) {
        var valuesType = undefined;
        var mustSet = false;
        var result = typeBuilder.getLazyMapType(function () {
            mustSet = true;
            return valuesType;
        });
        setTypeForPath(path, result);
        path = path.push({ kind: PathElementKind.AdditionalProperty });
        valuesType = toType(additional, path, TypeNames_2.singularizeTypeNames(typeAttributes));
        if (mustSet) {
            result.deref()[0].setValues(valuesType);
        }
        return result;
    }
    function fromTypeName(schema, path, typeAttributes, typeName) {
        // FIXME: We seem to be overzealous in making attributes.  We get them from
        // our caller, then we make them again here, and then we make them again
        // in `makeClass`, potentially in other places, too.
        typeAttributes = makeAttributes(schema, TypeNames_2.modifyTypeNames(typeAttributes, function (tn) { return Support_1.defined(tn).makeInferred(); }));
        switch (typeName) {
            case "object":
                var required = void 0;
                if (schema.required === undefined) {
                    required = [];
                }
                else {
                    required = checkStringArray(schema.required);
                }
                var classType = undefined;
                if (schema.properties !== undefined) {
                    classType = makeClass(path, typeAttributes, Support_1.checkStringMap(schema.properties), required);
                }
                var mapType = undefined;
                if (schema.additionalProperties !== undefined) {
                    var additional = schema.additionalProperties;
                    if (additional === true) {
                        return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
                    }
                    else if (additional === false) {
                        if (classType === undefined) {
                            classType = makeClass(path, typeAttributes, {}, required);
                        }
                    }
                    else {
                        mapType = makeMap(path, typeAttributes, Support_1.checkStringMap(additional));
                    }
                }
                if (classType !== undefined && mapType !== undefined) {
                    return UnifyClasses_1.unifyTypes(immutable_1.Set([classType.deref()[0], mapType.deref()[0]]), typeAttributes, typeBuilder, true, true, conflateNumbers);
                }
                if (classType !== undefined)
                    return classType;
                if (mapType !== undefined)
                    return mapType;
                return typeBuilder.getMapType(typeBuilder.getPrimitiveType("any"));
            case "array":
                if (schema.items !== undefined) {
                    path = path.push({ kind: PathElementKind.Items });
                    return typeBuilder.getArrayType(toType(Support_1.checkStringMap(schema.items), path, TypeNames_2.singularizeTypeNames(typeAttributes)));
                }
                return typeBuilder.getArrayType(typeBuilder.getPrimitiveType("any"));
            case "boolean":
                return typeBuilder.getPrimitiveType("bool");
            case "string":
                if (schema.format !== undefined) {
                    switch (schema.format) {
                        case "date":
                            return typeBuilder.getPrimitiveType("date");
                        case "time":
                            return typeBuilder.getPrimitiveType("time");
                        case "date-time":
                            return typeBuilder.getPrimitiveType("date-time");
                        default:
                            return Support_1.panic("String format " + schema.format + " not supported");
                    }
                }
                return typeBuilder.getStringType(typeAttributes, undefined);
            case "null":
                return typeBuilder.getPrimitiveType("null");
            case "integer":
                return typeBuilder.getPrimitiveType("integer");
            case "number":
                return typeBuilder.getPrimitiveType("double");
            default:
                return Support_1.panic("not a type name: " + typeName);
        }
    }
    function convertToType(schema, path, typeAttributes) {
        typeAttributes = makeAttributes(schema, typeAttributes);
        function convertOneOrAnyOf(cases, kind) {
            if (!Array.isArray(cases)) {
                return Support_1.panic("oneOf or anyOf is not an array: " + cases);
            }
            // FIXME: This cast shouldn't be necessary, but TypeScript forces our hand.
            var types = cases.map(function (t, index) { return toType(Support_1.checkStringMap(t), path.push({ kind: kind, index: index }), typeAttributes).deref()[0]; });
            return UnifyClasses_1.unifyTypes(immutable_1.OrderedSet(types), typeAttributes, typeBuilder, true, true, conflateNumbers);
        }
        if (schema.$ref !== undefined) {
            var _a = parseRef(schema.$ref), ref = _a[0], refName_1 = _a[1];
            var _b = lookupRef(schema, path, ref), target = _b[0], targetPath = _b[1];
            var attributes = TypeNames_2.modifyTypeNames(typeAttributes, function (tn) {
                if (!Support_1.defined(tn).areInferred)
                    return tn;
                return new TypeNames_1.TypeNames(immutable_1.OrderedSet([refName_1]), immutable_1.OrderedSet(), true);
            });
            return toType(target, targetPath, attributes);
        }
        else if (Array.isArray(schema.enum)) {
            var cases = schema.enum;
            var haveNull = cases.indexOf(null) >= 0;
            cases = cases.filter(function (c) { return c !== null; });
            var tref = typeBuilder.getEnumType(typeAttributes, immutable_1.OrderedSet(checkStringArray(cases)));
            if (haveNull) {
                return typeBuilder.getUnionType(typeAttributes, immutable_1.OrderedSet([tref, typeBuilder.getPrimitiveType("null")]));
            }
            else {
                return tref;
            }
        }
        else if (schema.type !== undefined) {
            var jsonTypes = checkTypeList(schema.type);
            if (jsonTypes.size === 1) {
                return fromTypeName(schema, path, typeAttributes, Support_1.defined(jsonTypes.first()));
            }
            else {
                var types = jsonTypes.map(function (n) { return fromTypeName(schema, path, typeAttributes, n).deref()[0]; });
                return UnifyClasses_1.unifyTypes(types, typeAttributes, typeBuilder, true, true, conflateNumbers);
            }
        }
        else if (schema.oneOf !== undefined) {
            return convertOneOrAnyOf(schema.oneOf, PathElementKind.OneOf);
        }
        else if (schema.anyOf !== undefined) {
            return convertOneOrAnyOf(schema.anyOf, PathElementKind.AnyOf);
        }
        else {
            return typeBuilder.getPrimitiveType("any");
        }
    }
    function toType(schema, path, typeAttributes) {
        // FIXME: This fromJS thing is ugly and inefficient.  Schemas aren't
        // big, so it most likely doesn't matter.
        var immutablePath = makeImmutablePath(path);
        var maybeType = typeForPath.get(immutablePath);
        if (maybeType !== undefined) {
            return maybeType;
        }
        var result = convertToType(schema, path, typeAttributes);
        setTypeForPath(immutablePath, result);
        return result;
    }
    references.forEach(function (topLevelRef, topLevelName) {
        var _a = lookupRef(root, exports.rootRef, topLevelRef), target = _a[0], targetPath = _a[1];
        var t = toType(target, targetPath, TypeNames_2.makeNamesTypeAttributes(topLevelName, false));
        typeBuilder.addTopLevel(topLevelName, t);
    });
}
exports.addTypesInSchema = addTypesInSchema;
function definitionRefsInSchema(rootJson) {
    if (typeof rootJson !== "object")
        return immutable_1.Map();
    var definitions = rootJson.definitions;
    if (typeof definitions !== "object")
        return immutable_1.Map();
    return immutable_1.Map(Object.keys(definitions).map(function (name) {
        return [name, exports.rootRef.push({ kind: PathElementKind.Definition, name: name })];
    }));
}
exports.definitionRefsInSchema = definitionRefsInSchema;
